from unittest import TestCase
from iprint.iprint import iprint, get_styled_text, status_print
import functools
import textwrap
import re


class TestIPrint(TestCase):
    def setUp(self):
        class FileMock:
            content = ""

            def write(self, content: str):
                self.content = content

        self.file_mock = FileMock()
        self.mprint_mock = functools.partial(iprint, file=self.file_mock)

    def test_write_in_file(self):
        self.mprint_mock('hello')
        actual = self.file_mock.content
        expected = 'hello\n'
        self.assertEqual(actual, expected)

    def test_end(self):
        self.mprint_mock('hello ', end="world")
        actual = self.file_mock.content
        expected = 'hello world'
        self.assertEqual(actual, expected)

    def test_multi_input(self):
        self.mprint_mock('hello', "world")
        actual = self.file_mock.content
        expected = 'hello, world\n'
        self.assertEqual(actual, expected)

    def test_sep(self):
        self.mprint_mock('hello', "world", sep=" test ")
        actual = self.file_mock.content
        expected = 'hello test world\n'
        self.assertEqual(actual, expected)


class TestGetStyledText(TestCase):
    def test_simple_string(self):
        actual = get_styled_text('hello', 4)
        expected = 'hello'
        self.assertEqual(actual, expected)

    def test_simple_list(self):
        actual = get_styled_text(['hello'], 4)
        expected = textwrap.dedent("""\
            [
                hello
            ]"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_simple_list_two_index(self):
        actual = get_styled_text(['hello', 'world'], 4)
        expected = textwrap.dedent("""\
            [
                hello,
                world
            ]"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_nested_two_dimensional_list(self):
        actual = get_styled_text([["hello"]], 4)
        expected = textwrap.dedent("""\
            [
                [
                    hello
                ]
            ]"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_nested_three_dimensional_list(self):
        actual = get_styled_text([[["hello"]]], 4)
        expected = textwrap.dedent("""\
            [
                [
                    [
                        hello
                    ]
                ]
            ]"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_nested_two_dimensional_list_two_index(self):
        actual = get_styled_text([["hello", "world"]], 4)
        expected = textwrap.dedent("""\
            [
                [
                    hello,
                    world
                ]
            ]"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_two_list_in_a_list(self):
        actual = get_styled_text([["hello", "world"], ["test", "case"]], 4)
        expected = textwrap.dedent("""\
            [
                [
                    hello,
                    world
                ],
                [
                    test,
                    case
                ]
            ]"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_string_list_in_a_list(self):
        actual = get_styled_text([["hello", "world"], "test"], 4)
        expected = textwrap.dedent("""\
            [
                [
                    hello,
                    world
                ],
                test
            ]"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_simple_dict(self):
        actual = get_styled_text({"hello": "world"}, 4)
        expected = textwrap.dedent("""\
            {
                hello:
                    world
            }"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_nested_dict(self):
        actual = get_styled_text({"test": {"hello": "world"}}, 4)
        expected = textwrap.dedent("""\
            {
                test:
                    {
                        hello:
                            world
                    }
            }"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_list_in_a_dict(self):
        actual = get_styled_text({"hello": ["world"]}, 4)
        expected = textwrap.dedent("""\
            {
                hello:
                    [
                        world
                    ]
            }"""
        )
        self.assertMultiLineEqual(actual, expected)

    def test_simple_set(self):
        actual = get_styled_text({"hello", "world"}, 4)
        expected = [
            textwrap.dedent("""\
                {
                    hello,
                    world
                }"""
            ),
            textwrap.dedent("""\
                {
                    world,
                    hello
                }"""
            )
        ]
        self.assertIn(actual, expected)

    def test_multi_line_text(self):
        actual = get_styled_text(["""line_1\nline_2"""], 4)
        expected = textwrap.dedent("""\
            [
                line_1
                line_2
            ]"""
        )
        self.assertIn(actual, expected)


class TestStatusPrint(TestCase):
    def setUp(self):
        class FileMock:
            content = ""

            def write(self, content: str):
                self.content = content

        self.file_mock = FileMock()
        self.status_print_mock = functools.partial(status_print, file=self.file_mock)

    def test_document(self):
        class Data:
            """this is a document"""

        self.status_print_mock(Data())
        actual = self.file_mock.content
        expected = textwrap.indent(textwrap.dedent("""\
            document:
                this is a document,"""
        ), ' ' * 4)
        self.assertIn(expected, actual)

    def test_instance_variable(self):
        class Data:
            def __init__(self, data):
                self.instance_variable = data

        self.status_print_mock(Data("instance_variable_value"))
        actual = self.file_mock.content
        expected = textwrap.indent(textwrap.dedent("""\
            instance_variables:
                {
                    instance_variable:
                        instance_variable_value
                },"""
        ), ' ' * 4)
        self.assertIn(expected, actual)

    def test_class_variable(self):
        class Data:
            class_variable = "class_variable_value"

        self.status_print_mock(Data())
        actual = self.file_mock.content
        expected = textwrap.indent(textwrap.dedent("""\
            class_variables:
                {
                    class_variable:
                        class_variable_value
                },"""
        ), ' ' * 4)
        self.assertIn(expected, actual)

    def test_methods(self):
        class Data:
            def instance_method(self):
                pass

        self.status_print_mock(Data())
        actual = self.get_normalize_memory_address(self.file_mock.content)
        expected = textwrap.indent(textwrap.dedent("""\
            methods:
                {
                    instance_method:
                        <function TestStatusPrint.test_methods.<locals>.Data.instance_method at <MEMORY_ADDRESS>>
                },"""
        ), ' ' * 4)

        self.assertIn(expected, actual)

    def test_class_method(self):
        class Data:
            @classmethod
            def class_method(cls):
                pass

        self.status_print_mock(Data())
        actual = self.get_normalize_memory_address(self.file_mock.content)
        expected = textwrap.indent(textwrap.dedent("""\
            class_methods:
                {
                    class_method:
                        <classmethod(<function TestStatusPrint.test_class_method.<locals>.Data.class_method at <MEMORY_ADDRESS>>)>
                },"""
        ), ' ' * 4)
        self.assertIn(expected, actual)

    def test_static_method(self):
        class Data:
            @staticmethod
            def static_method():
                pass

        self.status_print_mock(Data())
        actual = self.get_normalize_memory_address(self.file_mock.content)
        expected = textwrap.indent(textwrap.dedent("""\
            static_methods:
                {
                    static_method:
                        <staticmethod(<function TestStatusPrint.test_static_method.<locals>.Data.static_method at <MEMORY_ADDRESS>>)>
                },"""
        ), ' ' * 4)
        self.assertIn(expected, actual)

    def test_property(self):
        class Data:
            @property
            def property_variable(self):
                return "property_variable_value"

        self.status_print_mock(Data())
        actual = self.get_normalize_memory_address(self.file_mock.content)
        expected = textwrap.indent(textwrap.dedent("""\
            properties:
                {
                    property_variable:
                        <property object at <MEMORY_ADDRESS>>
                },"""
        ), ' ' * 4)
        self.assertIn(expected, actual)
    @staticmethod
    def get_normalize_memory_address(actual):
        return re.sub(r'at 0x[0-9a-fA-F]+', 'at <MEMORY_ADDRESS>', actual)



from abc import ABC, abstractmethod


class ToStr(ABC):
    def __init__(self, data, indent, text_level=0, colorize=False):
        self._data = data
        self._indent = indent
        self._text_level = text_level
        self._colorize = colorize

    @abstractmethod
    def execute(self) -> str:
        pass

    @property
    def indent_space(self):
        return " " * (self._indent * self._text_level)

    def indenter(self, data: list, open_icon: str, close_icon: str) -> str:
        output_str = f"{self.indent_space}{open_icon}\n"

        for item in data:
            output_str += get_styled_text(item, self._indent, self._text_level + 1, self._colorize)
            if item != data[-1]:
                output_str += ","
            output_str += "\n"

        output_str += f"{self.indent_space}{close_icon}"

        return output_str


class TupleToStr(ToStr):
    def execute(self) -> str:
        return self.indenter(self._data, "(", ")")


class ListToStr(ToStr):
    def execute(self) -> str:
        return self.indenter(self._data, "[", "]")


class DictToStr(ToStr):
    def execute(self) -> str:
        return self.indenter(self._data, "{", "}")

    def indenter(self, data: dict, open_icon: str, close_icon: str) -> str:
        output_str = f"{self.indent_space}{open_icon}\n"

        for key, value in data.items():
            output_str += get_styled_text(key, self._indent, self._text_level + 1, self._colorize)
            output_str += ":\n"
            output_str += get_styled_text(value, self._indent, self._text_level + 2, self._colorize)
            if key != list(data.keys())[-1]:
                output_str += ","
            output_str += "\n"

        output_str += f"{self.indent_space}{close_icon}"

        return output_str


class SetToStr(ToStr):
    def execute(self) -> str:
        return self.indenter(list(self._data), "{", "}")


class IntToStr(ToStr):
    def execute(self) -> str:
        return f"{self.indent_space}{str(self._data)}"


class StringToStr(ToStr):
    def get_output_text(self):
        new_data = self._data.replace("\n", f"\n{self.indent_space}")
        return f"{self.indent_space}{new_data}"

    def execute(self) -> str:
        return self.get_output_text()


class FloatToStr(ToStr):
    def execute(self) -> str:
        return f"{self.indent_space}{str(self._data)}"


class AnotherToStr(ToStr):
    def execute(self) -> str:
        return f"{self.indent_space}{str(self._data)}"

datatypes = {
    list: ListToStr,
    tuple: TupleToStr,
    dict: DictToStr,
    set: SetToStr,
    str: StringToStr,
    float: FloatToStr,
    int: IntToStr,
}
default_type = AnotherToStr


def get_styled_text(text, indent, text_level=0, colorize=False):
    if colorize:
        from color_conversions import datatypes, default_type
    else:
        global datatypes
        global default_type

    return datatypes.get(type(text), default_type)(text, indent, text_level, colorize).execute()




from abc import ABC, abstractmethod


class Color(ABC):
    @abstractmethod
    def get_color_code(self):
        pass


class CBlack(Color):
    def get_color_code(self):
        return "\033[1;30m"


class CRed(Color):
    def get_color_code(self):
        return "\033[1;31m"


class CGreen(Color):
    def get_color_code(self):
        return "\033[1;32m"


class CBlue(Color):
    def get_color_code(self):
        return "\033[1;34m"


class CMagenta(Color):
    def get_color_code(self):
        return "\033[1;35m"


class CCyan(Color):
    def get_color_code(self):
        return "\033[1;36m"


class CWhite(Color):
    def get_color_code(self):
        return "\033[1;37m"


class COrange(Color):
    def get_color_code(self):
        return "\033[1;33m"


class CDefault(Color):
    def get_color_code(self):
        return "\033[0m"


class ColorError(Exception):
    pass


def get_colored_message(message: str, color: Color, clear_color: Color = CDefault()):
    if not isinstance(color, Color) or not isinstance(clear_color, Color):
        raise ColorError("invalid color or clear color (need to be Color class like 'CWhite')")
    return f"{color.get_color_code()}{message}{clear_color.get_color_code()}"




from conversions import *
from tools.colored_text import get_colored_message, CBlue, CMagenta, CGreen, CCyan, COrange, CBlack, CRed, CWhite

list_color = CBlue()
tuple_color = CMagenta()
dict_color = CMagenta()
set_color = CBlue()
str_color = CGreen()
float_color = CRed()
int_color = COrange()
another_color = CWhite()


class ColorListToStr(ListToStr):
    def execute(self) -> str:
        open_icon = get_colored_message("[", list_color)
        close_icon = get_colored_message("]", list_color)
        return self.indenter(self._data, open_icon, close_icon)


class ColorTupleToStr(TupleToStr):
    def execute(self) -> str:
        print(self._data)
        open_icon = get_colored_message("(", tuple_color)
        close_icon = get_colored_message(")", tuple_color)
        return self.indenter(self._data, open_icon, close_icon)


class ColorDictToStr(DictToStr):
    def execute(self) -> str:
        open_icon = get_colored_message("{", dict_color)
        close_icon = get_colored_message("}", dict_color)
        return self.indenter(self._data, open_icon, close_icon)


class ColorSetToStr(SetToStr):
    def execute(self) -> str:
        open_icon = get_colored_message("{", set_color)
        close_icon = get_colored_message("}", set_color)
        return self.indenter(list(self._data), open_icon, close_icon)


class ColorFloatToStr(FloatToStr):
    def execute(self) -> str:
        return f"{self.indent_space}{get_colored_message(str(self._data), float_color)}"


class ColorIntToStr(IntToStr):
    def execute(self) -> str:
        return f"{self.indent_space}{get_colored_message(str(self._data), int_color)}"


class ColorStringToStr(StringToStr):
    def execute(self) -> str:
        return get_colored_message(self.get_output_text(), str_color)


class ColorAnotherToStr(AnotherToStr):
    def execute(self) -> str:
        return f"{self.indent_space}{get_colored_message(str(self._data), another_color)}"


datatypes = {
    list: ColorListToStr,
    tuple: ColorTupleToStr,
    dict: ColorDictToStr,
    set: ColorSetToStr,
    str: ColorStringToStr,
    float: ColorFloatToStr,
    int: ColorIntToStr,
}
default_type = ColorAnotherToStr



from setuptools import setup, find_packages

setup(
    name='indent_print',
    version='1.2.1',
    description='library can print better readable code',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author='matin ahmadi',
    author_email='matinahmadi.programmer@gmail.com',
    url='https://github.com/matinprogrammer/IPrint',
    packages=find_packages(),
    py_modules=['iprint', 'color_conversions', 'conversions'],
    install_requires=[

    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6'
)



import sys
import inspect
from conversions import get_styled_text


def iprint(*text, sep=", ", end="\n", indent=4, file=sys.stdout):
    out_put = ""
    for item in text:
        out_put += get_styled_text(item, indent)
        if item != text[-1]:
            out_put += sep
    file.write(out_put + end)


def cprint(*text, sep=" ", end="\n", indent=4):
    out_put = ""
    for item in text:
        out_put += get_styled_text(item, indent, colorize=True)
        if item != text[-1]:
            out_put += sep
    sys.stdout.write(out_put + end)


def status_print(instance, *, end="\n", indent=4, file=sys.stdout, colorize=False, show_dunder_attr=False):
    document = instance.__doc__
    try:
        size = instance.__sizeof__()
    except TypeError:
        size = sys.getsizeof (instance)
    try:
        instance_variables = dict(instance.__dict__)
    except AttributeError:
        instance_variables = {}

    class_variables = {}
    methods = {}
    class_methods = {}
    static_methods = {}
    properties = {}
    other_members = {}

    for name, attr in vars(instance.__class__).items():
        if isinstance(attr, staticmethod):
            static_methods[name] = attr
        elif isinstance(attr, classmethod):
            class_methods[name] = attr
        elif isinstance(attr, property):
            properties[name] = attr
        elif not name.startswith('__') and not name.endswith('__'):
            if inspect.ismethod(attr) or inspect.isfunction(attr) or inspect.ismethoddescriptor(attr):
                methods[name] = attr
            else:
                class_variables[name] = attr
        else:
            if name != "__doc__":
                other_members[name] = attr

    data = {
        "document": document,
        "size": size,
        "instance_variables": instance_variables,
        "class_variables": class_variables,
        "methods": methods,
        "class_methods": class_methods,
        "static_methods": static_methods,
        "properties": properties,
        "other_members": other_members if show_dunder_attr else "'off'",
    }

    if colorize:
        if file != sys.stdout:
            raise RuntimeError("can't write colorized in file")
        cprint(data, end=end, indent=indent)
    else:
        iprint(data, end=end, indent=indent, file=file)


if __name__ == "__main__":
    # Example:
    class AnotherData:
        """document of test class"""
        class_variable = "class_variable_value"

        def __init__(self, data):
            self.instance_variable = data

        def instance_method(self):
            pass

        @classmethod
        def class_method(cls):
            pass

        @staticmethod
        def static_method():
            pass

        @property
        def property_variable(self):
            return "property_variable_value"


    string_data = "my name is matin"
    int_data = 20
    float_data = 1.55
    another_data = AnotherData("instance_variable_value")
    dict_data = {"author": "matin ahmadi", "github": "https://github.com/matinprogrammer"}
    set_data = {1, 2, 3}
    tuple_data = (1, 2)
    list_data = [string_data, int_data, float_data, another_data, dict_data, set_data, tuple_data,  [[["test list"]]]]

    print("result of mprint: ")
    iprint(list_data)

    print("\nresult of cprint: ")
    cprint(list_data)

    print("\nresult of status_print: ")
    status_print(another_data, show_dunder_attr=True, colorize=True)
